/*
  Macro           Tally
  Author          Carlo Hogeveen
  Website         eCarlo.nl/tse
  Compatibility   TSE 4.0 upwards, all variants
  Version         v1.1   30 Apr 2023

  This tool uses a TSE search-like input, and lists in a new unnamed buffer
  how many times each match of the search string was found.

  When using regular expressions the tool becomes awsome!
  Where a standard TSE search finds the lines in which matches occur,
  Tally finds the matches themselves, and lists each unique one with their
  count.

  Example usages:
    List and count all unique ip addresses, email addresses or URLs.
    List and count all unique values of the Nth field of a log or .csv file.
    List and count all unique usages of syntax in software programs.
    List and count all unique words, characters, names or abbreviations.


  PARAMETERS

  All TSE search strings can be used, with this nonsensical exception:
    Using "\c" in a regular expression might cause matches to be overlooked.

  All TSE's search options can be used, with these differences:
    "g" is redundant: It is always used whether you specified it or not.
    "v" is redundant: It is always used whether you specified it or not.

    "i" (lower-case): Searches and tallies case-insensitive.
    "I" (upper-case): Searches case-insensitive, tallies case-sensitive.

    "x" will not ignore the "w" option like TSE's Find does.
    "x" will not find overlapping results like TSE's Find does:
        "[a-z]#" will not find "cat", "at" and "t" for the same "cat".


  NOTES

  Tally will close (not unmark) an open block before searching,
  regardless of whether the "l" option is used.

  In the output, matches containing spaces will be quoted.


  INSTALLATION

  Copy this file to TSE's "mac" directory and compile it there, for example by
  opening it there and applying the Macro Compile menu.

  As a tool, "Tally" is a good candidate to add to TSE's Potpourri menu.


  RUN

  Run the tool by executing "Tally" from the Macro Execute menu.
  Or run it from the Potpourri menu if you configured it there.


  USAGE TIPS

  Regular expressions are awesome!
    Searching for "{cat}|{dog}" with options "gixw" will give you a tally for
    both "cat" and "dog". The "w" option prevents the search from tallying
    "catastrophe".

    Searching for "cat.@" with options "giwx" will find "cat", "cats" and
    "catastrophe", and tallies each word separately.

  How to tally all words
    An obvious use case would be to tally all words in a file.
    A solution is to search for TSE's WordSet characters enclosed between "["
    and "]#" with search options "giwx".
    You can copy TSE's WordSet characters from this menu:
      Options, Full Configuration, Command/Format Options, WordSet Characters.
    A default TSE installation's WordSet is "0-9A-Z_a-z".
    So there tallying all words would be done by searching for "[0-9A-Z_a-z]#"
    with search options "gixw".

  Note about case-insensitivity
    If you search and tally case-insensitive, then results are tallied together
    under the lower-cased version of the result.
    Case-insensitive searching and tallying probably reflects Tally's typical
    use.


  HISTORY

  v1         1 Apr 2023
    Initial release.

  v1.0.1     1 Apr 2023
    A tiny cosmetic improvement.

  v1.1      30 Apr 2023
    Search option "a" was not implemented, but I forgot to disable it,
    causing an infinite loop if a user used it.
    Now search option "a" is implemented and searches all files.

    Made search option "g" redundant too.
    Now it is always used, whether you specify it or not.

    The Options prompt is now more self-explanatory.

    Improved the documentation.
*/



// Global constants and semi-constants
#define CCF_OPTIONS                    _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_
string  MACRO_NAME    [MAXSTRINGLEN] = ''
string  SEARCH_OPTIONS_FOOTER   [75] = 'i = Always ignore case.  I = Search case-insensitive, tally case-sensitive.'
string  SEARCH_OPTIONS_QUESTION [63] = 'Options [ALiIWX] (All-files bLock i/I-gnore-case Word reg-eXp):'
string  VALID_SEARCH_OPTIONS     [8] = '^acliwx$'


string proc standardize_options(string dirty_options)
  string  clean_options [11] = ''
  integer i                  = 0
  string  option         [1] = ''

  for i = 1 to Length(dirty_options)
    option = Lower(dirty_options[i])
    if      Pos(option, VALID_SEARCH_OPTIONS)
    and not Pos(option, clean_options)
      clean_options = clean_options + option
    endif
  endfor

  return(clean_options)
end standardize_options


/*
  The following function determines the minimum amount of characters that a
  given regular expression can match. For the practical purpose that they are
  valid matches too, "^" and "$" are pretended to have length 1.
  The purpose of this procedure is, to be able to beforehand avoid searching
  for an empty string, which is logically pointless because it always succeeds
  (with one exception: past the end of the last line).
  Searching for an empty string in a loop makes it infinate and hangs a macro.
  Properly applied, using this procedure can avoid that.
*/
integer proc minimum_regexp_length(string s)
  integer addition           = 0
  integer i                  = 0
  integer NEXT_TIME          = 2
  integer orred_addition     = 0
  integer prev_addition      = 0
  integer prev_i             = 0
  integer result             = 0
  integer tag_level          = 0
  integer THIS_TIME          = 1
  integer use_orred_addition = FALSE
  // For each character.
  for i = 1 to Length(s)
    // Ignore this zero-length "character".
    if Lower(SubStr(s,i,2)) == "\c"
      i = i + 1
    else
      // Skip index for all these cases so they will be counted as one
      // character.
      case SubStr(s,i,1)
        when "["
          while i < Length(s)
          and   SubStr(s,i,1) <> "]"
            i = i + 1
          endwhile
        when "\"
          i = i + 1
          case Lower(SubStr(s,i,1))
            when "x"
              i = i + 2
            when "d", "o"
              i = i + 3
          endcase
      endcase
      // Now start counting.
      if use_orred_addition == NEXT_TIME
         use_orred_addition =  THIS_TIME
      endif
      // Count a literal character as one:
      if SubStr(s,i-1,1) == "\" // (Using the robustness of SubStr!)
        addition = 1
      // Count a tagged string as the length of its subexpression:
      elseif SubStr(s,i,1) == "{"
        prev_i = i
        tag_level = 1
        while i < Length(s)
        and   (tag_level <> 0 or SubStr(s,i,1) <> "}")
          i = i + 1
          if SubStr(s,i,1) == "{"
            tag_level = tag_level + 1
          elseif SubStr(s,i,1) == "}"
            tag_level = tag_level - 1
          endif
        endwhile
        addition = minimum_regexp_length(SubStr(s,prev_i+1,i-prev_i-1))
      // For a "previous character or tagged string may occur zero
      // times" operator: since it was already counted, subtract it.
      elseif SubStr(s,i,1) in "*", "@", "?"
        addition = -1 * Abs(prev_addition)
      // This is a tough one: the "or" operator.
      // For now subtract the length of the previous character or
      // tagged string, but remember doing so, because you might have
      // to add it again instead of the length of the character or
      // tagged string after the "or" operator.
      elseif SubStr(s,i,1) == "|"
        addition           = -1 * Abs(prev_addition)
        orred_addition     = Abs(prev_addition)
        use_orred_addition = NEXT_TIME
      else
      // Count ordinary characters as 1 character.
        addition = 1
      endif
      if use_orred_addition == THIS_TIME
        if orred_addition < addition
          addition = orred_addition
        endif
        use_orred_addition = FALSE
      endif
      result        = result + addition
      prev_addition = addition
    endif
  endfor
  return(result)
end minimum_regexp_length


proc display_options_footer()
  UnHook(display_options_footer)
  WindowFooter(SEARCH_OPTIONS_FOOTER)
end display_options_footer


proc WhenLoaded()
  MACRO_NAME = SplitPath(CurrMacroFilename(), _NAME_)
end WhenLoaded


proc Main()
  integer first_found_buffer_id         = 0
  integer first_found_line              = 0
  integer first_found_pos               = 0
  integer found_next                    = FALSE
  integer i                             = 0
  integer num_tally_tokens              = 0
  integer ok                            = TRUE
  string  search_options           [11] = ''
  string  search_options_entered   [11] = ''
  string  search_options_next      [11] = ''
  string  search_options_start     [11] = ''
  string  search_string  [MAXSTRINGLEN] = ''
  integer tally_case_insensitive        = FALSE
  integer tally_count                   = 0
  integer tally_counters_id             = 0
  integer tally_regular_expressions     = FALSE
  string  tally_token    [MAXSTRINGLEN] = ''
  integer tally_tokens_id               = 0
  integer tally_words                   = FALSE

  PushLocation()
  tally_counters_id = CreateTempBuffer()
  ChangeCurrFilename(MACRO_NAME + ':TallyCountersBuffer', CCF_OPTIONS)
  tally_tokens_id   = CreateTempBuffer()
  ChangeCurrFilename(MACRO_NAME + ':TallyTokensBuffer'  , CCF_OPTIONS)
  PopLocation()

  ok = Ask('Search and tally for:', search_string, _FIND_HISTORY_)

  if  ok
  and search_string == ''
    Warn('You cannot search for an empty string.')
    ok = FALSE
  endif

  if ok
    Hook(_PROMPT_STARTUP_, display_options_footer)
    if Ask(Format(SEARCH_OPTIONS_QUESTION:
                  Max(Length(SEARCH_OPTIONS_QUESTION),
                      Length(SEARCH_OPTIONS_FOOTER  )) * -1),
           search_options_entered,
           _FINDOPTIONS_HISTORY_)
      search_options       = standardize_options(search_options_entered)
      search_options_start = search_options + 'g'
      search_options_next  = search_options
    else
      ok = FALSE
    endif
  endif

  if  ok
  and Pos('x', search_options)
  and minimum_regexp_length(search_string) == 0
    Warn('Your regular expression matches an empty string.')
    ok = FALSE
  endif

  if ok
    // Close (not unmark) an open block.
    if Query(Marking)
      case isBlockInCurrFile()
        when _INCLUSIVE_
          MarkStream()
        when _NON_INCLUSIVE_
          MarkChar()
        when _LINE_
          MarkLine()
        when _COLUMN_
          MarkColumn()
      endcase
    endif

    // Start searching.
    tally_case_insensitive    = Pos('i', search_options_entered) // Lower-case!
    tally_regular_expressions = Pos('x', search_options)
    tally_words               = Pos('w', search_options)

    PushLocation()

    if lFind(search_string, search_options_start)
      first_found_buffer_id = GetBufferId()
      first_found_line      = CurrLine()
      first_found_pos       = CurrPos()

      repeat
        tally_token = iif(tally_case_insensitive,
                          Lower(GetFoundText()), GetFoundText())
        // In exceptional cases tally_token can already be empty.
        // Make "x" not overrule "w":
        if  tally_regular_expressions
        and tally_words
        and not EquiStr(GetWord(), tally_token)
          tally_token = ''
        endif
        if tally_token <> ''
          tally_count = GetBufferInt(tally_token, tally_counters_id)
          if tally_count == 0
            num_tally_tokens = num_tally_tokens + 1
            ok = SetBufferStr(Str(num_tally_tokens), tally_token, tally_tokens_id)
            ok = ok and SetBufferInt(tally_token, 1, tally_counters_id)
          else
            ok = SetBufferInt(tally_token, tally_count + 1, tally_counters_id)
          endif
          Right(Length(tally_token))
        endif
        if ok
          found_next = (    lFind(search_string, search_options_next)
                        and (  GetBufferId() <> first_found_buffer_id
                            or CurrLine()    <> first_found_line
                            or CurrPos()     <> first_found_pos))
        endif
      until not ok
         or not found_next
    endif

    if not ok
      Warn(MACRO_NAME; 'error: SetBuffer...() error: Is TSE out of memory?')
    endif

    PopLocation()
  endif

  if ok
    if num_tally_tokens
      NewFile()
      for i = 1 to num_tally_tokens
        tally_token = GetBufferStr(Str(i), tally_tokens_id)
        tally_count = GetBufferInt(tally_token, tally_counters_id)
        AddLine(Format(tally_count:10;;; QuotePath(tally_token)))
      endfor
      BegFile()
    else
      Warn('No matches found.')
      ok = FALSE
    endif
  endif

  AbandonFile(tally_counters_id)
  AbandonFile(tally_tokens_id)
  PurgeMacro(MACRO_NAME)
end Main

